||| Djot inline formatting parser
module SSG.Djot.Inline

import Control.Eff
import Data.List1
import Data.Maybe
import Data.String
import Derive.Prelude
import Derive.Pretty

-- Just for iutils unit tests
import System

%language ElabReflection

--******************************
--*       Data Structures      *
--******************************

||| Types of inline styling
public export
data Inline : Type where
  Text : (text : String) -> Inline
  SoftLineBreak : Inline

%runElab derive "Inline" [Eq, Show, Pretty]

--******************************
--*      Parsing Utilities     *
--******************************

||| Type alias for inline parsing
IParser : Type -> Type
IParser t = Eff [State (List Char), Fail] t

||| Get the next char, modifiying the state
popChar : IParser Char
popChar = do
  x :: xs <- get
    | [] => fail
  put xs
  pure x

||| Get the next char, without modifying the state
peekChar : IParser Char
peekChar = do
  x :: xs <- get
    | [] => fail
  pure x

||| Attempt to parse something without propagating the failure
try : IParser t -> IParser (Maybe t)
try x = do
  state <- get
  x <- lift . runFail $ x
  case x of
    Nothing => do
      put state
      pure Nothing
    Just y => pure $ Just y


||| Choose a parser
oneOf : List (IParser t) -> IParser t
oneOf [] = fail
oneOf (x :: xs) = do
  state <- get
  x <- lift . runFail $ x
  case x of
    Nothing => do
      put state
      oneOf xs
    Just y => pure y

||| Parse 0+ of something
many : IParser t -> IParser (List t)
many x = do
  Just y <- try x
    | _ => pure []
  map (y ::) $ many x

||| Parse 1+ of something
atLeastOne : IParser t -> IParser (List1 t)
atLeastOne x = do
  Just y <- try x
    | _ => fail
  map (y :::) (many x)

||| Match exactly the given string
exactly : String -> IParser (List Char)
exactly str = exactly' (unpack str)
  where
    exactly' : List Char -> IParser (List Char)
    exactly' [] = pure []
    exactly' (x :: xs) = do
      y <- popChar
      if x == y
        then map (x ::) (exactly' xs)
        else fail

||| Match vertical whitespace
vert : IParser (List Char)
vert = oneOf
  [ exactly "\n"
  , exactly "\r\n"
  , exactly "\r"
  ]

||| Match horizontal whitespace
horiz : IParser (List Char)
horiz = oneOf
  [ exactly " "
  , exactly "\t"
  ]

||| Run a parser
runIParser : (str : String) -> IParser t -> Maybe t
runIParser str x = 
  fst . extract . runState (unpack str) . runFail $ x

--******************************
--*           Syntax           *
--******************************

-- Forward declare so we can get mutually recursive
||| Top level parser function for Inline Content
pInline : IParser Inline

||| Parse a character as plain text
text : IParser Inline
text = do
  c <- popChar
  pure $ Text (singleton c)

||| Parse a soft linebreak
softLineBreak : IParser Inline
softLineBreak = do
  _ <- atLeastOne singleBreak
  pure SoftLineBreak
  where
    singleBreak : IParser ()
    singleBreak = do
      _ <- many horiz
      _ <- vert
      _ <- many horiz
      pure ()

-- Definition of pInline
pInline = oneOf
  [ softLineBreak
  , text
  ]

--******************************
--*       Post Processing      *
--******************************

||| Combine adjacent Text entries
combineTexts : List Inline -> List Inline
combineTexts [] = []
combineTexts (Text a :: (Text b :: xs)) =
  combineTexts (Text (a ++ b) :: xs)
combineTexts (x :: xs) = x :: combineTexts xs

||| Remove trailing soft line breaks
removeTrailingSoftbreak : List Inline -> List Inline
removeTrailingSoftbreak xs = reverse (removeTrailingSoftbreak' (reverse xs))
  where
    removeTrailingSoftbreak' : List Inline -> List Inline
    removeTrailingSoftbreak' (SoftLineBreak :: xs) = removeTrailingSoftbreak' xs
    removeTrailingSoftbreak' xs = xs

||| Top level post processor
postProcess : List Inline -> List Inline
postProcess xs = removeTrailingSoftbreak . combineTexts $ xs

--******************************
--* Top Level Parsing Function *
--******************************

||| Parse a string as inline content
export
inline : (input : String) -> List Inline
inline input =
  postProcess . fromMaybe [] . runIParser input $ many pInline

--******************************
--*         Unit Tests         *
--******************************

golden : (input : String) -> (reference : List Inline) -> IO Bool
golden input reference = do
  putStrLn "Input: \{show input}"
  let opts = Opts 78
  let ref_pretty = Doc.render opts $ pretty reference
  putStrLn "Reference:\n\{unlines . map ("  " ++) . lines $ ref_pretty}"
  let output = inline input
  let out_pretty = Doc.render opts $ pretty output
  putStrLn "Output:\n\{unlines . map ("  " ++) . lines $ out_pretty}"
  pure $ reference == output

-- @@test Just text parses as text
testTextAsText : IO Bool
testTextAsText =
  golden "Hello World!" [Text "Hello World!"]

-- @@test Soft line break smoke
smokeSoftLineBreak : IO Bool
smokeSoftLineBreak =
  golden "Hello\nWorld!" [Text "Hello", SoftLineBreak, Text "World!"]

-- @@test Soft line break double line break
testDoubleSoftLineBreak : IO Bool
testDoubleSoftLineBreak =
  golden "Hello\n\nWorld!" [Text "Hello", SoftLineBreak, Text "World!"]

-- @@test Soft line break trailing line break
testTrailingSoftLineBreak : IO Bool
testTrailingSoftLineBreak =
  golden "Hello\n\nWorld!\n" [Text "Hello", SoftLineBreak, Text "World!"]

-- @@test Soft line break multiple trailing line breaks
testTrailingSoftLineBreaks : IO Bool
testTrailingSoftLineBreaks =
  golden "Hello\n\nWorld!\n\n\n" [Text "Hello", SoftLineBreak, Text "World!"]
